/**
 * Created by IntelliJ IDEA.
 * User: Luchesar
 * Date: May 2, 2004
 * Time: 10:14:20 AM
 * To change this template use Options | File Templates.
 */
package inet.warehouse.model.persistance.file;

import inet.warehouse.model.persistance.APersister;
import inet.warehouse.model.persistance.ObjectNotFoundException;
import inet.warehouse.model.persistance.Id;
import inet.warehouse.model.persistance.PersistenceException;
import inet.warehouse.model.AModel;

import java.io.*;
import java.util.*;
import java.beans.PropertyDescriptor;
import java.lang.reflect.Array;
import java.nio.channels.FileChannel;
import java.nio.channels.FileLock;
import java.nio.ByteBuffer;

import org.apache.commons.beanutils.PropertyUtils;
import junit.framework.TestCase;

public class FilePersister extends APersister {
    private static FilePersister mInstance;
    private Thread mSaveThread;
    private BeanComparator defaultSort = new BeanComparator(new String[]{"id"});


    //all tables
    private HashMap modelMapping;
    public static final String DB_FILE_KEY = "dbFilePath";

    private class Save extends Thread {
        public void run() {
            while (!interrupted()) {
                try {
                    sleep(10 * 1000 * 60);
                } catch (InterruptedException e) {
                    break;
                }
                try {
                    saveInFile();
                } catch (IOException e) {
                    e.printStackTrace();
                    continue;
                }
            }
        }
    }

    private FilePersister() {
        try {
            loadFromFile();
            mSaveThread = new Save();
            mSaveThread.setPriority(Thread.NORM_PRIORITY - 2);
            mSaveThread.setDaemon(true);
            mSaveThread.start();
            Runtime.getRuntime().addShutdownHook(new Thread() {
                public void run() {
                    try {
                        saveInFile();
                    } catch (IOException e) {
                        e.printStackTrace();
                    }
                }
            });
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    public static FilePersister getI() {
        if (mInstance == null) {
            mInstance = new FilePersister();
        }
        return mInstance;
    }

    public AModel create(Class aClass) throws PersistenceException {
        try {
            return (AModel) aClass.newInstance();
        } catch (Exception e) {
            throw new PersistenceException(e);
        }
    }

    public void save(AModel aModel) throws PersistenceException {
        new TraverceSaver().save(aModel);
    }


    public void delete(Class aClass, Id aId) throws ObjectNotFoundException {
        Map modelMap = getModelMap(aClass);
        if (modelMap.containsKey(aId)) {
            modelMap.remove(aId);
        } else {
            throw new ObjectNotFoundException("object not found");
        }
    }

    public void deleteAll(Class aClass) throws PersistenceException {
        modelMapping.remove(aClass.getName());
    }

    public AModel load(Class aClass, Id aId) throws ObjectNotFoundException, PersistenceException {
        return new TraverceLoader().load(aClass, aId);
    }

    public AModel[] listAll(Class aClass) throws PersistenceException {
        return sort(new TraverceLoader().listAll(aClass));
    }

    public AModel[] listBy(Class aClass, String aPropertyName, Object aPropertyValue) throws PersistenceException {
        return sort(new TraverceLoader().listBy(aClass, aPropertyName, aPropertyValue));
    }

    public AModel[] listBy(Class aClass, Map propertyValueMapping) throws PersistenceException {
        return sort(new TraverceLoader().listBy(aClass, propertyValueMapping));
    }

    public AModel[] listAll(Class aClass,String[] orderBy) throws PersistenceException {
        return sort(new TraverceLoader().listAll(aClass),orderBy);
    }

    public AModel[] listBy(Class aClass, String aPropertyName, Object aPropertyValue,String[] orderBy) throws PersistenceException {
        return sort(new TraverceLoader().listBy(aClass, aPropertyName, aPropertyValue),orderBy);
    }

    public AModel[] listBy(Class aClass, Map propertyValueMapping,String[] orderBy) throws PersistenceException {
        return sort(new TraverceLoader().listBy(aClass, propertyValueMapping),orderBy);
    }

    protected Map getModelMap(Class aClass) {
        Map result = null;
        if (!modelMapping.containsKey(aClass.getName())) {
            result = new HashMap();
            modelMapping.put(aClass.getName(), new HashMap());
        } else {
            result = (Map) modelMapping.get(aClass.getName());
        }
        return result;
    }

    public void saveInFile() throws IOException {
        
        String filePath = System.getProperty(DB_FILE_KEY);
        filePath = filePath == null ? "db.dat" : filePath;
        File f = new File(filePath);
        f.createNewFile();

        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        ObjectOutputStream out = new ObjectOutputStream(baos);
        out.writeObject(modelMapping);
        out.close();


        FileOutputStream fout = new FileOutputStream(f);
        try {
            FileChannel fch = fout.getChannel();
            FileLock lock = null;
            try {
                lock = fch.lock();
                try {
                    fch.write(ByteBuffer.wrap(baos.toByteArray()));
                } finally {
                    lock.release();
                }
            } finally {
                fch.close();
            }
        } finally {
            fout.close();
        }

    }

    protected void loadFromFile() throws IOException, Exception {
        String filePath = System.getProperty(DB_FILE_KEY);
        filePath = filePath == null ? "db.dat" : filePath;
        File f = new File(filePath);
        if (!f.exists()) {
            modelMapping = new HashMap();
        } else {
            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            FileInputStream fin = new FileInputStream(f);
            try {
                FileChannel fch = fin.getChannel();
                try {
                    ByteBuffer buffer = ByteBuffer.allocate((int)fch.size());
                    for (int i = -1 ;(i = fch.read(buffer)) != -1;) {
                        baos.write(buffer.array(),0,i);
                        buffer.clear();
                    }
                } finally {
                    fch.close();
                }
            } finally {
                fin.close();
            }
            ObjectInputStream in = new ObjectInputStream(new ByteArrayInputStream(baos.toByteArray()));
            modelMapping = (HashMap) in.readObject();
        }
    }

    public void finalize() {
        mSaveThread.interrupt();
    }

    private class TraverceSaver {
        private final HashSet mCache = new HashSet();

        public void save(AModel aModel) throws PersistenceException {
            if (aModel == null || mCache.contains(aModel.getId())) {
                return;
            }

            mCache.add(aModel.getId());

            Map modelMap = getModelMap(aModel.getClass());
            PropertyDescriptor[] descriptors = PropertyUtils.getPropertyDescriptors(aModel.getClass());
            try {
                for (int i = 0; i < descriptors.length; i++) {
                    PropertyDescriptor descriptor = descriptors[i];

                    if (descriptor.getWriteMethod() != null) {
                        Object value = PropertyUtils.getSimpleProperty(aModel, descriptor.getName());
                        String name = descriptor.getName();

                        if (value instanceof AModel[]) {
                            AModel[] models = (AModel[]) value;
                            for (int j = 0; j < models.length; j++) {
                                if (models[j] == null) continue;
                                String parentPropertyName = getPropertyNameWithType(aModel.getClass(), models[j].getClass());
                                if (parentPropertyName == null) {
                                    throw new PersistenceException("The detail property with type->" + models[j].getClass() + "<- must have property for the Master type->" + aModel.getClass() + "<-");
                                }

                                PropertyUtils.setSimpleProperty(models[j], parentPropertyName, aModel);
                                save(models[j]);
                            }
                        } else {
                            Map properties = null;
                            if (modelMap.containsKey(aModel.getId())) {
                                properties = (Map) modelMap.get(aModel.getId());
                            } else {
                                properties = new HashMap();
                                modelMap.put(aModel.getId(), properties);
                            }
                            if (AModel.class.isAssignableFrom(descriptor.getPropertyType())) {
                                if (value == null) continue;
                                Id id = ((AModel) value).getId();
                                properties.put(name, id);
//                                save((AModel) value);
                            } else {
                                properties.put(name, value);
                            }
                        }
                    }
                }
            } catch (Exception e) {
                throw new PersistenceException(e);
            }
        }

        private boolean containPropertyWithType(Class aParent, Class aChild) {
            PropertyDescriptor[] properties = PropertyUtils.getPropertyDescriptors(aChild);
            for (int k = 0; k < properties.length; k++) {
                if (aParent.equals(properties[k].getPropertyType()))
                    return true;
            }
            return false;
        }
    }

    private class TraverceLoader {
        private final HashMap mCache = new HashMap();

        public AModel load(Class aClass, Id aId) throws ObjectNotFoundException, PersistenceException {
            if (mCache.containsKey(aId)) {
                return (AModel) mCache.get(aId);
            }

            Map modelMap = getModelMap(aClass);
            try {
                if (modelMap.containsKey(aId)) {
                    Map savedProperties = (Map) modelMap.get(aId);
                    AModel model = (AModel) aClass.newInstance();
                    mCache.put(aId, model);
                    PropertyDescriptor[] descriptors = PropertyUtils.getPropertyDescriptors(aClass);

                    for (int i = 0; i < descriptors.length; i++) {
                        PropertyDescriptor descriptor = descriptors[i];
                        if (descriptor.getWriteMethod() != null) {
                            String name = descriptor.getName();

                            Object value = null;
                            if (AModel.class.isAssignableFrom(descriptor.getPropertyType())) {
                                Id id = (Id) savedProperties.get(name);
                                if (id != null) {
                                    try {
                                        value = load(descriptor.getPropertyType(), id);
                                    } catch (ObjectNotFoundException e) {
                                        //object was deleted
                                    }
                                }
                            } else if (AModel[].class.isAssignableFrom(descriptor.getPropertyType())) {
                                value = listBy(descriptor.getPropertyType().getComponentType(),
                                        getPropertyNameWithType(aClass, descriptor.getPropertyType().getComponentType()), aId);
                            } else {
                                value = savedProperties.get(name);
                            }
                            PropertyUtils.setSimpleProperty(model, name, value);
                        }
                    }
                    return model;
                }
            } catch (Exception e) {
                e.printStackTrace();
                throw new PersistenceException(e);
            }
            throw new ObjectNotFoundException("object not found");
        }

        public AModel[] listAll(Class aClass) throws PersistenceException {
            Map modelMap = getModelMap(aClass);
            ArrayList all = new ArrayList(modelMap.size());
            try {
                for (Iterator i = modelMap.entrySet().iterator(); i.hasNext();) {
                    Map.Entry e = (Map.Entry) i.next();
                    all.add(load(aClass, (Id) e.getKey()));
                }
            } catch (Exception e) {
                e.printStackTrace();
                throw new PersistenceException(e);
            }
            return (AModel[]) all.toArray((Object[]) Array.newInstance(aClass, 0));
        }


        public AModel[] listBy(Class aClass, String aPropertyName, Object aPropertyValue) throws PersistenceException {
            Map modelMap = getModelMap(aClass);
            ArrayList finded = new ArrayList();
            try {
                for (Iterator i = modelMap.entrySet().iterator(); i.hasNext();) {
                    Map.Entry entry = (Map.Entry) i.next();
                    Map properties = (Map) entry.getValue();
                    if (properties.containsKey(aPropertyName)) {
                        Object propertyValue = properties.get(aPropertyName);
                        if (propertyValue.equals(aPropertyValue)) {
                            finded.add(load(aClass, (Id) entry.getKey()));
                        }
                    }
                }
            } catch (Exception e) {
                e.printStackTrace();
                throw new PersistenceException(e);
            }
            return (AModel[]) finded.toArray((Object[]) Array.newInstance(aClass, 0));
        }

        public AModel[] listBy(Class aClass, Map propertyValueMapping) throws PersistenceException {
            Map modelMap = getModelMap(aClass);
            ArrayList finded = new ArrayList();
            try {
                for (Iterator i = modelMap.entrySet().iterator(); i.hasNext();) {
                    Map.Entry entry = (Map.Entry) i.next();
                    Map properties = (Map) entry.getValue();
                    int matches = 0;
                    ;
                    for (Iterator keys = propertyValueMapping.keySet().iterator(); keys.hasNext();) {
                        String key = (String) keys.next();
                        if (properties.containsKey(key)) {
                            Object propertyValue = properties.get(key);
                            if (propertyValue.equals(propertyValueMapping.get(key))) {
                                matches++;
                            } else {
                                break;
                            }
                        }
                    }
                    if (propertyValueMapping.size() == matches) {
                        finded.add(load(aClass, (Id) entry.getKey()));
                    }
                }

            } catch (Exception e) {
                e.printStackTrace();
                throw new PersistenceException(e);
            }
            return (AModel[]) finded.toArray((Object[]) Array.newInstance(aClass, 0));
        }
    }

    private String getPropertyNameWithType(Class aParent, Class aChild) {
        PropertyDescriptor[] properties = PropertyUtils.getPropertyDescriptors(aChild);
        for (int k = 0; k < properties.length; k++) {
            if (aParent.getName().equals(properties[k].getPropertyType().getName()))
                return properties[k].getName();
        }
        return null;
    }

    private AModel[] sort(AModel[] aModel) {
        Arrays.sort(aModel,defaultSort);
        return aModel;
    }

    private AModel[] sort(AModel[] aModel,String[] properties) {
        Arrays.sort(aModel,new BeanComparator(properties));
        return aModel;
    }


}
