package inet.warehouse.web;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.util.Vector;
import java.io.UnsupportedEncodingException;

/**
 * Base class for all action classes
 */
public abstract class Action {

    public String executeAction(HttpServletRequest request, HttpServletResponse response) throws ActionException {
        try {
            request.setCharacterEncoding(Constants.DEFAULT_CHAR_SET);
        } catch (UnsupportedEncodingException e) {
            throw new ActionException(e);
        }
        Vector errors = new Vector();
        request.setAttribute(Constants.VAR_ERRORS, errors);

        String result = null;
        if (shouldCheckAuthentication()) {
            // check if user is logged in
            if (request.getSession().getAttribute(Constants.AUTHENTICATED_USER_SESSION_KEY) == null) {
                addError(request, "errors.notLoggedIn");
                result = "/";
            } else {
                result = execute(request, response);
            }
        } else {
            result = execute(request, response);
        }

        return result;
    }

    /**
     * An action should override this method to allow unauthenticated execution
     * @return
     */
    protected boolean shouldCheckAuthentication() {
        return true;
    }

    /**
     * The actual
     * @param request
     * @param response
     * @return
     * @throws ActionException
     */
    protected abstract String execute(HttpServletRequest request, HttpServletResponse response) throws ActionException;


    public static final void addError(HttpServletRequest request, String errorMessage) {
        addError(request,new Error(errorMessage,null));
    }

    public static final void addError(HttpServletRequest request, String errorMessage,Object[] params) {
        addError(request,new Error(errorMessage,params));
    }

    public static final void addError(HttpServletRequest request, Error error) {
        Vector errors = (Vector) request.getAttribute(Constants.VAR_ERRORS);
        if (errors == null) {
            errors = new Vector();
        }
        errors.add(error);
    }
}
